from dataclasses import dataclass
from .data_classes import OSTIRResult

from .ostir_calculations import calc_dG_standby_site, calc_longest_loop_bulge, calc_dG_mRNA_rRNA, calc_dG_mRNA, calc_expression_level, calc_longest_helix


# Default values. @TODO: Find a way to overwrite these
auto_dangles = True
dangles_default = "all"
start_codon_energies = {"ATG": -1.194, "AUG": -1.194, "GTG": -0.0748, "GUG": -0.0748, "TTG": -0.0435,
                                        "UUG": -0.0435, "CTG": -0.03406, "CUG": -0.03406}  # hybridization to CAT
standby_site_length = 4
decimal_places = 4
cutoff = 35

def ostir_worker(task, start_position, task_ID, verbosity=0):
    """ Run OSTIR on a single sequence. """

    # Get the start codon from the start position and the task
    sequence = task.sequence.upper()
    start_codon = start_position[1]
    start_position = start_position[0]

    # Set dangles based on length between 5' end of mRNA and start codon
    if auto_dangles:
        global cutoff
        cutoff = int(cutoff)
        if start_position > cutoff:
            dangles = "none"

        else:
            dangles = "all"

    else:
        dangles = dangles_default

    # Start codon energy
    dG_start_codon = start_codon_energies[start_codon]

    # Energy of mRNA folding. This also gets us the kinetic score
    dG_mRNA, mRNA_structure, kinetic_score, min_bp_prob = calc_dG_mRNA(sequence, start_position, dangles, task.constraints)

    # Energy of mRNA:rRNA hybridization & folding
    dg_mRNA_rRNA_output = calc_dG_mRNA_rRNA(sequence, task.rrna, start_position, dangles, task.constraints)
    dG_mRNA_rRNA_withspacing = dg_mRNA_rRNA_output[0]
    mRNA_rRNA_structure = dg_mRNA_rRNA_output[1]
    spacing_value = dg_mRNA_rRNA_output[2]
    if not dG_mRNA_rRNA_withspacing:
        return None

    dG_mRNA_rRNA_withspacing -= 2.481  # Modifying hybridization penalty to match NUPACK

    dG_mRNA_rRNA_nospacing = mRNA_rRNA_structure["dG_mRNA_rRNA"]
    dG_mRNA_rRNA_nospacing -= 2.481  # Modifying hybridization penalty to match NUPACK

    # Standby site correction:
    dG_standby_site, corrected_structure = calc_dG_standby_site(mRNA_rRNA_structure,
                                                                            dangles,
                                                                            standby_site_length,
                                                                            task.constraints,
                                                                            task.rrna)

    # Total energy is mRNA:rRNA + start - rRNA - mRNA - standby_site
    dG_total = dG_mRNA_rRNA_withspacing + dG_start_codon - dG_mRNA - dG_standby_site

    # Calculate dG to open SD sequence
    # ddG_SD_open = self.calc_dG_SDopen(mRNA_structure, mRNA_rRNA_structure)

    loop_result = calc_longest_loop_bulge(structure=mRNA_structure)
    helical_loop_list, bulge_loop_list = loop_result[0], loop_result[1]

    parallel_output = OSTIRResult(
        name= task.name,
        start_codon=start_codon,
        start_position=start_position+1,
        expression=round(calc_expression_level(dG_total),
                            decimal_places),
        RBS_distance_bp=int(spacing_value),
        dG_total=round(float(dG_total), decimal_places),
        dG_rRNA_mRNA=round(float(dG_mRNA_rRNA_nospacing), decimal_places),
        dG_mRNA=round(float(dG_mRNA), decimal_places),
        dG_spacing=round(float(mRNA_rRNA_structure["dG_spacing"]), decimal_places),
        dG_standby=round(float(dG_standby_site), decimal_places),
        dG_start_codon=round(float(dG_start_codon), decimal_places),

        mRNA_structure=mRNA_structure,
        mRNA_rRNA_uncorrected_structure=mRNA_rRNA_structure,
        mRNA_rRNA_corrected_structure=corrected_structure,
        longest_helix=calc_longest_helix(mRNA_structure),
        dG_start_energy=dG_start_codon,
        helical_loop=helical_loop_list,
        min_bp_prob=min_bp_prob,
        bulge_loop=bulge_loop_list,
        kinetic_score=kinetic_score
    )


    return task_ID, parallel_output


"""

Some of the codebase of this file originated from the Salis Lab's RBS calculator which is distributed under GPL3.
See <http://www.gnu.org/licenses/>.
Copyright 2008-2009 is owned by the University of California Regents. All rights reserved.
"""

from shutil import which
from operator import itemgetter
from dataclasses import dataclass
from functools import cache
import warnings
import os
import array
import sys
import re
import RNA


#  On import check dependencies
dependencies = [which('RNAfold') is not None,
                which('RNAsubopt') is not None,
                which('RNAeval') is not None]
if False in dependencies:
    warnings.warn('RBS Calculator Vienna is missing dependency ViennaRNA!')
#  End dependency check

debug = 0

@dataclass
class ViennaConstants():
    debug = 0
    RT = 0.61597
    material = 'rna2004'

vienna_constants = ViennaConstants()


#Class that encapsulates all of the functions from ViennaRNA
class ViennaRNA(dict):


    def __init__(self, Sequence_List):
        self.RT = 0.61597 #gas constant times 310 Kelvin (in units of kcal/mol)

        for seq in Sequence_List:
            if re.compile('[ATGCU]', re.IGNORECASE).match(seq) is None:
                error_string = "Invalid letters found in inputted sequences. Only ATGCU allowed. \n Sequence is \"" + str(seq) + "\"."
                raise ValueError(error_string)

        self["sequences"] = Sequence_List


def mfe(sequences, constraints, temp , dangles):
    '''
    Calculate the MFE of a sequence using ViennaRNA as a module
    '''

    params = get_paramater_object(vienna_constants.material, temp, dangles)
    seq_string = "&".join(sequences).upper().replace("T", "U")

    rna = RNA.fold_compound(seq_string, params)

    if constraints:
        rna.hc_add_from_db(constraints)

    vienna_mfe = rna.mfe()
    bracket_string = vienna_mfe[0]
    vienna_energy = round(vienna_mfe[1], 2)

    bp_x, bp_y = convert_bracket_to_numbered_pairs(bracket_string)

    mfe_basepairing_x = array.array('i', bp_x)
    mfe_basepairing_y = array.array('i', bp_y)
    mfe_energy = vienna_energy

    return mfe_basepairing_x, mfe_basepairing_y, mfe_energy


def subopt(sequences, constraints, energy_gap, temp = 37.0, dangles = "some"):

    #self["subopt_composition"] = strands

    if temp <= 0: raise ValueError("The specified temperature must be greater than zero.")

    seq_string = "&".join(sequences)
    #seq_string = self["sequences"][0]

    #print(f'SEQ STRING: {seq_string}')

    params = get_paramater_object(vienna_constants.material, temp, dangles)
    seq_string = "&".join(sequences).upper().replace("T", "U")

    rna = RNA.fold_compound(seq_string, params)

    if constraints:
        if len(sequences) > 1:
            constraints = constraints + "."*len(sequences[1])
        rna.hc_add_from_db(constraints)

    vienna_subopt = rna.subopt(int((energy_gap+2.481)*100))
    subopt_output = [[str(output.structure), str(round(output.energy, 3))] for output in vienna_subopt]

    subopt_energy = []
    subopt_basepairing_x = []
    subopt_basepairing_y = []

    if len(sequences) > 1:
        subopt_output = process_fold_outputs(subopt_output, multi_sequence=True)
    else:
        subopt_output = process_fold_outputs(subopt_output, multi_sequence=False)

    for finding in subopt_output:
        
        subopt_energy.append(float(finding[1]))
        subopt_basepairing_x.append(finding[2])
        subopt_basepairing_y.append(finding[3])


    return subopt_energy, subopt_basepairing_x, subopt_basepairing_y


def energy(sequences, base_pairing_x, base_pairing_y, Temp, dangles): 
    if Temp <= 0: raise ValueError("The specified temperature must be greater than zero.")
    temp = Temp

    seq_string = "&".join(sequences)
    strands = [len(seq) for seq in sequences]
    bracket_string = convert_numbered_pairs_to_bracket(strands,base_pairing_x,base_pairing_y)

    params = get_paramater_object(vienna_constants.material, temp, dangles)
    seq_string = "&".join(sequences).upper().replace("T", "U")

    rna = RNA.fold_compound(seq_string, params)
    result = rna.eval_structure(bracket_string.replace("&", ""))

    vienna_energy = round(result, 2)
    return vienna_energy


def convert_numbered_pairs_to_bracket(strands, bp_x, bp_y):

    bp_x = [pos-1 for pos in bp_x[:]] #Shift so that 1st position is 0
    bp_y = [pos-1 for pos in bp_y[:]] #Shift so that 1st position is 0

    bracket_notation = []
    counter = 0
    for (strand_number, seq_len) in enumerate(strands):
        if strand_number > 0: bracket_notation.append("&")
        for pos in range(counter, seq_len+counter):
            if pos in bp_x:
                bracket_notation.append("(")
            elif pos in bp_y:
                bracket_notation.append(")")
            else:
                bracket_notation.append(".")
        counter += seq_len

    return "".join(bracket_notation)

def convert_bracket_to_numbered_pairs(bracket_string):

    #print(all_seq_len)
    bp_x = []
    bp_y = []

    for _ in range(bracket_string.count(")")):
        bp_y.append([])

    last_nt_x_list = []
    num_strands=0
    #print(bracket_string)
    for (pos,letter) in enumerate(bracket_string[:]):
        if letter == ".":
            pass

        elif letter == "(":
            bp_x.append(pos-num_strands)
            last_nt_x_list.append(pos-num_strands)

        elif letter == ")":
            nt_x = last_nt_x_list.pop() #nt_x is list of "(" except last entry
            #print('this is the last_nt_x_list ' + str(last_nt_x_list.pop()))
            nt_x_pos = bp_x.index(nt_x)
            bp_y[nt_x_pos] = pos-num_strands

        elif letter == "&":
            num_strands += 1

        else:
            print("Error! Invalid character in bracket notation.")

    if len(last_nt_x_list) > 0:
        print("Error! Leftover unpaired nucleotides when converting from bracket notation to numbered base pairs.")

    if len(bp_y) > 1:
        bp_x = [int(pos+1) for pos in bp_x[:]] #Shift so that 1st position is 1
        bp_y = [int(pos+1) for pos in bp_y[:]] #Shift so that 1st position is 1
        #print("subopt bp_x " + str(bp_x))
        #print("subopt bp_y " + str(bp_y))

    return bp_x, bp_y

def process_fold_outputs(findings, multi_sequence = False):
    filtered_findings = []
    # Earlier vienna versions would report 'folds' without any actual folding when running under WSL.
    # This code provides backwards compatibility for versions with that bug
    for finding in findings:
        if len(finding) != 2:
            continue
        if multi_sequence:
            binding_positions = finding[0]
            binding_positions = binding_positions.split("&")
            if ")" not in binding_positions[1] and "(" not in binding_positions[1]:
                #print("This binding site is fake")
                continue
            else:
                bp_x, bp_y = convert_bracket_to_numbered_pairs(finding[0])
                #print(strands, bp_x, bp_y)
                filtered_findings.append([finding[0], finding[1], bp_x, bp_y])
        else:
            bp_x, bp_y = convert_bracket_to_numbered_pairs(finding[0])
            filtered_findings.append([finding[0], finding[1], bp_x, bp_y])
    # Linux distributions of Vienna RNA tiebreak the sorting of outputs differently than Mac. This code aims to
    # make sorting of the output consistent, which rarely affects downstream predictions by sorting by energy with
    # the string of the fold as the tiebreaker
    findings_by_energy = {}
    for finding in filtered_findings:  # Sorts by energy
        if finding[1] in findings_by_energy.keys():
            findings_by_energy[finding[1]].append(finding)
        else:
            findings_by_energy[finding[1]] = [finding]
    energy_values = [[float(i), i] for i in findings_by_energy.keys()]  # real keys may have hanging 0's
    energy_values = sorted(energy_values, key=itemgetter(1))
    sorted_findings = []
    for value in energy_values:
        dict_index = value[1]
        partitioned_findings = findings_by_energy[str(dict_index)]
        partitioned_findings = sorted(partitioned_findings, key=itemgetter(0))
        sorted_findings = sorted_findings + partitioned_findings

    return sorted_findings

def get_paramater_file(parameter):
    if parameter == 'rna1999':
        filepath = os.path.dirname(sys.modules['ostir'].__file__)+'/rna_turner1999.par'
    elif parameter == 'rna2004':
        filepath = os.path.dirname(sys.modules['ostir'].__file__)+'/rna_turner2004.par'
    else:
        filepath = ''
    return filepath

@cache
def get_paramater_object(parametersfile, temp, dangles):
    filepath = get_paramater_file(parametersfile)
    RNA.params_load(filepath)

    params = RNA.md()
    params.noLP = 1
    params.noPS = 1
    params.temperature = temp

    if dangles == "none":
        params.dangles = 0
    elif dangles == "some":
        params.dangles = 1
    elif dangles == "all":
        params.dangles = 2

    return params


def clean_temp_file(file):
    if os.path.isfile(file):
        os.remove(file)



# cython: profile=True

import math
from dataclasses import dataclass
from copy import deepcopy
from .ViennaRNA import ViennaRNA, subopt, mfe, energy
import numpy as np
from datetime import datetime

@dataclass
class OstirConstants():
    Beta: float = 0.40002512
    RT_eff: float = 1/Beta
    logK: float = 7.279194329
    K: float = math.exp(logK)
    infinity = np.inf
    RNA_model: str = "rna2004"
    auto_dangles: bool = True
    dangles_default: str = "all"
    temp: float = 37.0
    optimal_spacing: int = 5  # aligned spacing
    cutoff: int = 35

        # From OSTIR calibration using Salis2009 data. See calibration directory for procedure
    dG_spacing_constant_push = np.array([17.20965071, 3.46341492, 1.790848365, 3.0], dtype=np.float64)
    dG_spacing_constant_pull = np.array([0.06422042, 0.275640836, 0.0], dtype=np.float64)
    cutoff: int = 35  # number of nt +- start codon considering for folding
    standby_site_length: int = 4  # Number of nt before SD sequence that must be unpaired for ribosome binding
    start_codons: list = ("ATG", # substituted U for T in actual calcs. Ignores CTG/CUG
                             "AUG",
                             "GTG",
                             "GUG",
                             "TTG",
                             "UUG")
    footprint: int = 1000
    energy_cutoff: float = 3.0
    verbose: bool = False

@dataclass
class StartEnergies:
    """Class to hold the start codon energies"""
    ATG: float = -1.194
    AUG: float = -1.194
    GTG: float = -0.0748
    GUG: float = -0.0748
    TTG: float = -0.0435
    UUG: float = -0.0435
    CTG: float = -0.03406
    CUG: float = -0.03406

ostir_constants = OstirConstants()
start_energies = StartEnergies()

# Timer decorator
def timer(func):
    def wrapper(*args, **kwargs):
        start = datetime.now()
        result = func(*args, **kwargs)
        end = datetime.now()
        print(f"[{datetime.now()}] {func.__name__} took {end - start}")
        return result
    return wrapper


def calc_longest_loop_bulge(structure, output_start_end=False, InRBSOnly=False, RBS=None):
    """Calculate the longest helical loop and bulge structure
    (longest contiguous list of un-base paired nucleotides starting and
    ending with a helix (loop -> same helix, bulge -> different helix) in the secondary structure"""

    mRNA = structure["mRNA"]

    bp_x = structure["bp_x"]
    bp_y = structure["bp_y"]

    loop_length = 0
    begin_helix = 1

    bulge_loop_list = []
    helical_loop_list = []
    end_helix = 0

    if output_start_end:
        bulge_loop_start_end = []
        helical_loop_start_end = []

    if InRBSOnly and RBS is not None:
        RBS_begin = mRNA.find(RBS)
        RBS_end = RBS_begin + len(RBS)
        loop_start = RBS_begin
        loop_end = RBS_end +1

    else:
        loop_start = 1
        loop_end = len(mRNA) + 1


    # Find loops. Find bulges.
    for i in range(loop_start, loop_end):
        if np.count_nonzero(bp_x == i) == 0 and np.count_nonzero(bp_y == i) == 0:  # nth nucleotide is not base-paired.
            # Determine if nearest neighbor nucleotides are base-paired
            x_1, x_2, y_1, y_2 = np.count_nonzero(bp_x == i - 1), np.count_nonzero(bp_x == i + 1), np.count_nonzero(bp_y == i - 1), np.count_nonzero(bp_y == i + 1)

            # print "#", n, (x_1,x_2,y_1,y_2)

            # middle unpaired nt
            if (x_1, x_2, y_1, y_2) == (0, 0, 0, 0):
                loop_length += 1

            # single mismatch -- loop
            elif (x_1, x_2, y_1, y_2) == (1, 0, 0, 1) or (x_1, x_2, y_1, y_2) == (0, 1, 1, 0):
                loop_length += 1
                begin_helix = i - 1
                end_helix = i + 1

            # single mismatch -- bulge
            elif (x_1, x_2, y_1, y_2) == (1, 1, 0, 0) or (x_1, x_2, y_1, y_2) == (0, 0, 1, 1):
                loop_length += 1
                begin_helix = i - 1
                end_helix = i + 1

            # starting unpaired nt
            elif (x_1, x_2, y_1, y_2) == (1, 0, 0, 0) or (x_1, x_2, y_1, y_2) == (0, 0, 1, 0):
                loop_length += 1
                begin_helix = i - 1

            # ending unpaired nt
            elif (x_1, x_2, y_1, y_2) == (0, 1, 0, 0) or (x_1, x_2, y_1, y_2) == (0, 0, 0, 1):
                loop_length += 1
                end_helix = i + 1

            # 1,0,1,0 is impossible w/o psuedoknots
            # 0,1,0,1 is impossible w/o psuedoknots
            # Also, all binary combinations with 3 or 4 true are impossible
            # (n-1 or n+1 can not be in both bp_x and bp_y)


        elif loop_length > 0:
            # Bulge or loop?
            # loop

            # print "begin = ", begin_helix
            # print "end = ", end_helix

            if (np.count_nonzero(bp_x == begin_helix) > 0 and np.count_nonzero(bp_y == end_helix) > 0 and np.where(bp_x == begin_helix) == np.where(bp_y ==
                    end_helix)):
                helical_loop_list.append(loop_length)
                loop_length = 0

                if output_start_end:
                    # Also return the starting and ending positions of each loop/bulge
                    helical_loop_start_end.append((begin_helix, end_helix))

            else:

                bp_end = 0
                bp_begin = 0

                if np.count_nonzero(bp_x == end_helix) > 0:
                    bp_begin = bp_y[np.where(bp_x == end_helix)]
                if np.count_nonzero(bp_y == end_helix) > 0:
                    bp_end = bp_x[np.where(bp_y == end_helix)]

                if np.count_nonzero(bp_x == begin_helix) > 0:
                    bp_end = bp_y[np.where(bp_x == begin_helix)]
                if np.count_nonzero(bp_y == begin_helix) > 0:
                    bp_begin = bp_x[np.where(bp_y == begin_helix)]

                if bp_end > bp_begin:
                    bulge_loop_list.append(loop_length)
                    loop_length = 0

                    if output_start_end:
                        # Also return the starting and ending positions of each loop/bulge
                        bulge_loop_start_end.append((begin_helix, end_helix))

                else:
                    loop_length = 0
    if output_start_end:
        return (helical_loop_list, bulge_loop_list, helical_loop_start_end, bulge_loop_start_end)
    else:
        return (helical_loop_list, bulge_loop_list)


def calc_longest_helix(structure: ViennaRNA):
    """Calculate the longest helical structure (longest contiguous list of base pairings)
    in the secondary structure"""

    bp_x = structure["bp_x"]
    bp_y = structure["bp_y"]

    longest_helix = 0
    helix_length = 1

    for (nt_x, nt_y) in zip(bp_x, bp_y):
        if (np.count_nonzero(bp_x == nt_x + 1) > 0 and np.count_nonzero(bp_y == nt_y - 1) > 0):
            helix_length += 1
        else:
            longest_helix = max(longest_helix, helix_length)
            helix_length = 1

    return longest_helix


def calc_kinetic_score(mRNA_in=None, bp_x_in=None, bp_y_in=None):
    """Calculate a "kinetic score", a heuristic measure of the maximum time required for
    the mRNA secondary structure to form. This is related to the RNA polymer model by David et. al.
    This heuristic should not be used in any way to quantify the folding kinetics of an mRNA sequence
    because it completely ignores cooperative RNA folding mechanisms, such as zipping or strand
    displacement. Here, we use it to eliminate mRNA sequences that MAY fold slowly."""


    mRNA = mRNA_in
    bp_x = bp_x_in
    bp_y = bp_y_in

    mrnalen = len(mRNA)
    largest_range_helix = 0

    for (nt_x, nt_y) in zip(bp_x, bp_y):
        if nt_x <= len(mRNA) and nt_y <= mrnalen:
            val = nt_y - nt_x
            largest_range_helix = max(val, largest_range_helix)

    kinetic_score = largest_range_helix / mrnalen
    if largest_range_helix > 0:
        min_bp_prob = largest_range_helix ** (-1.44)  # From David et. al.
    else:
        min_bp_prob = 1.0

    return kinetic_score, min_bp_prob


def calc_dG_standby_site(structure_old, dangles, standby_site_length, constraints, rRNA):
    """Calculates the dG_standby given the structure of the mRNA:rRNA complex

    To calculate the mfe structure while disallowing base pairing at the standby site,
    we split the folded mRNA sequence into three parts: (i) a pre-sequence (before the standby
    site) that can fold; (ii) the standby site, which can not fold; (iii) the 16S rRNA binding
    site and downstream sequence, which has been previously folded.
    """


    structure = deepcopy(structure_old)
    mRNA = structure["mRNA"]
    bp_x = structure["bp_x"]
    bp_y = structure["bp_y"]
    energy_before = structure["dG_mRNA_rRNA"]  # without spacing effects


    # Identify the most 5p mRNA nt that is bound to rRNA
    most_5p_mRNA = 0
    for (nt_x, nt_y) in zip(bp_x, bp_y):
        if nt_x <= len(mRNA) and nt_y > len(mRNA):  # nt_x is mRNA, nt_y is rRNA, they are bound.
            most_5p_mRNA = nt_x  # starts counting from 0
            break

    # Extract the base pairings that are 3' of the most_5p_mRNA base pairing
    bp_x_3p = []
    bp_y_3p = []
    for (nt_x, nt_y) in zip(bp_x, bp_y):
        if nt_x >= most_5p_mRNA:
            bp_x_3p.append(nt_x)
            bp_y_3p.append(nt_y)

    #Create the mRNA subsequence
    mRNA_subsequence = mRNA[0:max(0,most_5p_mRNA - standby_site_length - 1)]
    if constraints is None:
        constraint_subsequence = None
    else:
        constraint_subsequence = constraints[0:max(0,most_5p_mRNA - standby_site_length - 1)]

    # Fold it and extract the base pairings
    if (len(mRNA_subsequence)) > 0:
        mfe_basepairing_x, mfe_basepairing_y, _ = mfe([mRNA_subsequence], constraint_subsequence, temp=ostir_constants.temp, dangles=dangles)
        bp_x_5p = mfe_basepairing_x
        bp_y_5p = mfe_basepairing_y
    else:
        bp_x_5p = []
        bp_y_5p = []

    # Put the sets of base pairings together
    bp_x_after = []
    bp_y_after = []
    for (nt_x, nt_y) in zip(bp_x_5p, bp_y_5p):
        bp_x_after.append(nt_x)
        bp_y_after.append(nt_y)

    for (nt_x, nt_y) in zip(bp_x_3p, bp_y_3p):
        bp_x_after.append(nt_x)
        bp_y_after.append(nt_y)

    # Calculate its energy
    energy_after = energy([mRNA, rRNA], bp_x_after, bp_y_after, dangles=dangles, Temp=ostir_constants.temp)

    dG_standby_site = energy_before - energy_after
    if dG_standby_site > 0.0:
        dG_standby_site = 0.0
    # catch negative 0

    index = structure["MinStructureID"]


    structure["bp_x"] = np.array(bp_x_after, dtype=np.intc)
    structure["bp_y"] = np.array(bp_y_after, dtype=np.intc)
    structure["subopt_energy"][index] = energy_after
    structure["dG_mRNA_rRNA_corrected"] = energy_after

    return (dG_standby_site, structure)


def calc_dG_mRNA(mRNA, start_pos, dangles, constraints):
    """Calculates the dG_mRNA given the mRNA sequence."""
    mRNA = cutoff_mRNA(mRNA, start_pos)

    fold = ViennaRNA([mRNA])
    if constraints:
        constraints = constraints[max([0,start_pos-ostir_constants.cutoff]) : min([len(mRNA), start_pos+ostir_constants.cutoff])]
        mfe_basepairing_x, mfe_basepairing_y, mfe_energy = mfe([mRNA], constraints, temp=ostir_constants.temp, dangles=dangles)
    else:
        mfe_basepairing_x, mfe_basepairing_y, mfe_energy = mfe([mRNA], None, temp=ostir_constants.temp, dangles=dangles)

    structure = fold
    structure["mRNA"] = mRNA
    structure["bp_x"] = mfe_basepairing_x
    structure["bp_y"] = mfe_basepairing_y
    structure["dG_mRNA"] = mfe_energy
    structure["MinStructureID"] = 0

    dG_mRNA_folding = mfe_energy


    kinetic_score, min_bp_prob = calc_kinetic_score(mRNA, mfe_basepairing_x, mfe_basepairing_y)

    return dG_mRNA_folding, structure, kinetic_score, min_bp_prob



def calc_dG_mRNA_rRNA(mRNA_in, rRNA, start_pos, dangles, constraints):
    """Calculates the dG_mRNA_rRNA from the mRNA and rRNA sequence.
    Considers all feasible 16S rRNA binding sites and includes the effects of non-optimal spacing."""

    # Collect all constants
    cutoff = ostir_constants.cutoff
    temp = ostir_constants.temp
    energy_cutoff = ostir_constants.energy_cutoff

    begin = max(0, start_pos - cutoff)
    mRNA_len = min(len(mRNA_in), start_pos + cutoff)
    start_pos_in_subsequence = min(start_pos, cutoff)
    startpos_to_end_len = mRNA_len - start_pos_in_subsequence - begin


    # 1. identify a list of rRNA-binding sites. Binding sites are hybridizations between the mRNA and rRNA and can include mismatches, bulges, etc. Intra-molecular folding is also allowed within the mRNA. The subopt program is used to generate a list of optimal & suboptimal binding sites.
    # Constraints: the entire rRNA-binding site must be upstream of the start codon

    mRNA = mRNA_in[begin:start_pos]
    if begin == start_pos:
        raise ValueError("Warning: There is a leaderless start codon, which is being ignored.")

    #include viennaRNA folding constraints due to binding of global regulator


    if constraints and len(constraints) >= len(mRNA):
        constraints = constraints[begin:start_pos] #added so constraints file will be the same as the sequence
    elif constraints and len(constraints) < len(mRNA):
        constraints = constraints + "."*(len(mRNA)-len(constraints)) # Fill the rest with dots
    else:
        constraints = None
    subopt_energy, subopt_basepairing_x, subopt_basepairing_y = subopt([mRNA, rRNA], constraints , energy_cutoff, dangles=dangles, temp=temp)


    if len(subopt_basepairing_x) == 0:
        return None, None, None

    # 2. Calculate dG_spacing for each 16S rRNA binding site

    # Calculate the aligned spacing for each binding site in the list
    aligned_spacing = []
    loop_len = len(subopt_basepairing_x)
    for i in range(loop_len):
        aligned_spacing.append(calc_aligned_spacing(rRNA, mRNA, start_pos_in_subsequence, subopt_basepairing_x[i], subopt_basepairing_y[i]))

    dG_spacing_list = []
    dG_mRNA_rRNA = []
    dG_mRNA_rRNA_withspacing = []

    # Calculate dG_spacing using aligned spacing value. Add it to dG_mRNA_rRNA.
    for i in range(loop_len):
        dG_mRNA_rRNA.append(subopt_energy[i])
        val = calc_dG_spacing(aligned_spacing[i])
        dG_spacing_list.append(val)
        dG_mRNA_rRNA_withspacing.append(val + subopt_energy[i])

    # 3. Find 16S rRNA binding site that minimizes dG_spacing+dG_mRNA_rRNA.
    _, index = find_min(dG_mRNA_rRNA_withspacing)
    dG_spacing_final = dG_spacing_list[index]
    spacing_value = aligned_spacing[index]

    # Check: Is the dG spacing large compared to the energy gap? If so, this means the list of suboptimal 16S rRNA binding sites generated by subopt is too short.
    if dG_spacing_final > energy_cutoff:
        if ostir_constants.verbose:
            print("Warning: The spacing penalty is greater than the energy gap. dG (spacing) = ", dG_spacing_final)

    # 4. Identify the 5' and 3' ends of the identified 16S rRNA binding site. Create a base pair list.



    most_5p_mRNA = ostir_constants.infinity
    most_3p_mRNA = -ostir_constants.infinity

    # Generate a list of rRNA-mRNA base paired nucleotides
    bp_x_target = []
    bp_y_target = []

    bp_x = subopt_basepairing_x[index]
    bp_y = subopt_basepairing_y[index]
    loop_len = len(bp_x)


    for i in range(loop_len):
        nt_x = bp_x[i]
        nt_y = bp_y[i]
        if nt_y > len(mRNA):  # nt is rRNA
            index_match = bp_y.index(nt_y)
            most_5p_mRNA = min(most_5p_mRNA, bp_x[index_match])
            most_3p_mRNA = max(most_3p_mRNA, bp_x[index_match])
            bp_x_target.append(nt_x)
            bp_y_target.append(nt_y)


    """
    The rRNA-binding site is between the nucleotides at positions most_5p_mRNA and most_3p_mRNA
    Now, fold the pre-sequence, rRNA-binding-sequence and post-sequence separately. 
    Take their base pairings and combine them together. Calculate the total energy. 
    For secondary structures, this splitting operation is allowed.
    We postulate that not all of the post-sequence can form secondary structures. 
    Once the 30S complex binds to the mRNA, it prevents the formation of secondary 
    structures that are mutually exclusive with ribosome binding. We define self.footprint 
    to be the length of the 30S complex footprint. Here, we assume that the entire mRNA 
    sequence downstream of the 16S rRNA binding site can not form secondary structures.
    """


    begin = int(begin)
    most_5p_mRNA = int(most_5p_mRNA)
    most_3p_mRNA = int(most_3p_mRNA)

    mRNA_pre = mRNA_in[begin:begin + most_5p_mRNA - 1]
    post_window_end = mRNA_len + 1
    post_window_begin = min(start_pos + ostir_constants.footprint, post_window_end)  # Footprint
    post_window_end = mRNA_len + 1
    mRNA_post = mRNA_in[post_window_begin:post_window_end]

    total_bp_x = []
    total_bp_y = []

    if constraints:
        pre_constraints = constraints[begin:begin+most_5p_mRNA-1]
        post_constraints = constraints[post_window_begin:post_window_end]
    else:
        pre_constraints = None
        post_constraints = None

    # Calculate pre-sequence folding
    if len(mRNA_pre) > 0:
        mfe_basepairing_x, mfe_basepairing_y, _ = mfe([mRNA_pre], pre_constraints, temp=temp, dangles=dangles, )
        bp_x_pre = mfe_basepairing_x
        bp_y_pre = mfe_basepairing_y

    else:
        bp_x_pre = []
        bp_y_pre = []

    # Add pre-sequence base pairings to total base pairings
    total_bp_x.extend(bp_x_pre)
    total_bp_y.extend(bp_y_pre)


    # Add rRNA-binding site base pairings to total base pairings
    if startpos_to_end_len < cutoff:
        rRNA_offset = startpos_to_end_len
    else:
        rRNA_offset = startpos_to_end_len

    total_bp_x.extend(bp_x_target)
    for nt_y in bp_y_target:
        total_bp_y.append(nt_y + rRNA_offset)

    # Calculate post-sequence folding
    if len(mRNA_post) > 0:
        mfe_basepairing_x, mfe_basepairing_y, _ = mfe([mRNA_post], post_constraints, temp=temp, dangles=dangles)
        bp_x_post = mfe_basepairing_x
        bp_y_post = mfe_basepairing_y
    else:
        bp_x_post = []
        bp_y_post = []

    offset = 0  # Begins at zero
    offset = post_window_begin - begin
    for (nt_x, nt_y) in zip(bp_x_post, bp_y_post):
        total_bp_x.append(nt_x + offset)
        total_bp_y.append(nt_y + offset)


    mRNA = mRNA_in[begin:mRNA_len]
    fold = ViennaRNA([mRNA, rRNA])

    total_energy = energy([mRNA, rRNA], total_bp_x, total_bp_y, Temp=temp, dangles=dangles)

    total_energy_withspacing = total_energy + dG_spacing_final

    structure = fold
    #structure["program"] = "subopt"
    structure["mRNA"] = mRNA
    structure["MinStructureID"] = 0
    structure["dG_mRNA_rRNA"] = total_energy
    structure["dG_mRNA_rRNA_withspacing"] = total_energy_withspacing
    structure["dG_spacing"] = dG_spacing_final
    structure["subopt_energy"] = [total_energy_withspacing]
    structure["bp_x"] = np.array(total_bp_x, dtype=np.intc)
    structure["bp_y"] = np.array(total_bp_y, dtype=np.intc)

    return total_energy_withspacing, structure, spacing_value


def calc_dG_spacing(aligned_spacing):
    """Calculates the dG_spacing according to the value of the aligned spacing.
    This relationship was determined through experiments."""

    optimal_spacing = ostir_constants.optimal_spacing
    push_constants = ostir_constants.dG_spacing_constant_push
    pull_constants = ostir_constants.dG_spacing_constant_pull

    if aligned_spacing < optimal_spacing:
        ds = aligned_spacing - optimal_spacing

        dG_spacing_penalty = push_constants[0] / (
                    1.0 + math.exp(push_constants[1] * (ds + push_constants[2]))) ** \
                            push_constants[3]

    else:
        ds = aligned_spacing - optimal_spacing
        dG_spacing_penalty = pull_constants[0] * ds * ds + pull_constants[1] * ds + \
                            pull_constants[2]

    return dG_spacing_penalty


def calc_aligned_spacing(rRNA, mRNA, start_pos, bp_x, bp_y):
    """Calculates the aligned spacing between the 16S rRNA binding site and the start codon."""

    # rRNA is the concatenated at the end of the sequence in 5' to 3' direction
    # first: identify the farthest 3' nt in the rRNA that binds to the mRNA and return its mRNA base pairer

    rRNA_len = len(rRNA)
    Ok = False
    seq_len = len(mRNA) + rRNA_len

    loop_end = seq_len - rRNA_len

    for rRNA_nt in range(seq_len, loop_end, -1):

        if rRNA_nt in bp_y:
            rRNA_pos = bp_y.index(rRNA_nt)
            x_start = bp_x[rRNA_pos]
            if x_start < start_pos:
                Ok = True
                farthest_3_prime_rRNA = rRNA_nt - len(mRNA)

                mRNA_nt = bp_x[rRNA_pos]
                distance_to_start = start_pos - mRNA_nt + 1  # start_pos is counting starting from 0 (python)

                break
            else:
                break

    if Ok:
        aligned_spacing = distance_to_start - farthest_3_prime_rRNA
    else:
        aligned_spacing = ostir_constants.infinity

    return aligned_spacing


def calc_expression_level(dG):
    """Calculates the expression level of a given dG value."""
    K = ostir_constants.K
    RT_eff = ostir_constants.RT_eff
    return K * math.exp(-dG / RT_eff)


def find_min(input_list):
    """Finds the minimum of a list of numbers."""

    min_item = ostir_constants.infinity
    min_index = 0

    for i, item in enumerate(input_list):
        if item < min_item:
            min_item = item
            min_index = i

    return (min_item, min_index)


def find_start_codons(sequence, start_range):
    """Finds all start codons in an mRNA sequence. Creates a list."""

    seq_len = len(sequence)

    #Switch to zero-indexed positions happens here
    end_0 = min(start_range[1]-1, seq_len - 2)
    begin_0 = min(start_range[0]-1, end_0)

    for i in range(begin_0, end_0 + 1):
        codon = sequence[i:i + 3]
        if codon.upper() in ostir_constants.start_codons:
            yield (i, codon)
        else:
            pass

def cutoff_mRNA(mRNA, start_pos):
    mRNA = mRNA[max(0, start_pos - ostir_constants.cutoff):min(len(mRNA), start_pos + ostir_constants.cutoff)]
    return mRNA



#!/usr/bin/env python

import unittest
import os
import subprocess
import csv
import shutil

import ostir.ostir as ostir

THIS_DIR = os.path.dirname(os.path.abspath(__file__))
output_directory_path = os.path.join(THIS_DIR, 'output')


def csv_are_identical(csv_file_path_1, csv_file_path_2):
    'Utility function for comparing output, Returns whether files are identical as bool'
    failed = False
    try:
        with open(csv_file_path_1) as csv1:
            reader = csv.reader(csv1)
            csv_1_contents = [row for row in reader]
    except:
        print(f'Failed to open {str(csv_file_path_1)}')
        failed = True

    try:
        with open(csv_file_path_2) as csv2:
            reader = csv.reader(csv2)
            csv_2_contents = [row for row in reader]
    except:
        print(f'Failed to open {str(csv_file_path_2)}')
        failed = True

    if len(csv_1_contents) != len(csv_2_contents):
        print(f'CSVs are different lengths')
        failed = True

    differences = []
    for i in range(0, len(csv_1_contents)):
        if csv_1_contents[i] != csv_2_contents[i]:
            # print(csv_1_rows[i], "\n!=\n",csv_1_rows[i])
            differences.append(i)
    if differences:
        print(f'CSVs have {len(differences)} different values at rows {differences}')
        failed = True

    if failed == True:
        return False
    else:
        return True


##############################################################################################
# Unit tests (function calls)
##############################################################################################
class test_unit_run_ostir_one_sequence(unittest.TestCase):
    def test_unit_run_ostir_one_sequence(self):
        """
        Test run_ostir on one sequence
        """

        print("\n" + "Unit test: run_ostir_one_sequence")

        test_result = ostir.run_ostir("ACUUCUAAUUUAUUCUAUUUAUUCGCGGAUAUGCAUAGGAGUGCUUCGAUGUCAU")

        # print(test_result)

        expected_result = [
            {
                'name': 'unnamed',
                'start_codon': 'AUG',
                'start_position': 31,
                'expression': 2.1121,
                'RBS_distance_bp': 16,
                'dG_total': 16.3277,
                'dG_rRNA:mRNA': -0.481,
                'dG_mRNA': -7.2,
                'dG_spacing': 10.8027,
                'dG_standby': 0.0,
                'dG_start_codon': -1.194,
            },
            {
                'name': 'unnamed',
                'start_codon': 'GUG',
                'start_position': 41,
                'expression': 845.1532,
                'RBS_distance_bp': 8,
                'dG_total': 1.3491,
                'dG_rRNA:mRNA': -3.581,
                'dG_mRNA': -3.6,
                'dG_spacing': 1.4049,
                'dG_standby': 0.0,
                'dG_start_codon': -0.0748,
            },
            {
                'name': 'unnamed',
                'start_codon': 'AUG',
                'start_position': 49,
                'expression': 12448.1756,
                'RBS_distance_bp': 5,
                'dG_total': -5.375,
                'dG_rRNA:mRNA': -7.981,
                'dG_mRNA': -3.6,
                'dG_spacing': 0.0,
                'dG_standby': -0.2,
                'dG_start_codon': -1.194,
            }
        ]

        self.assertEqual(test_result, expected_result)


class test_unit_run_ostir_one_sequence_new_aSD(unittest.TestCase):
    def test_unit_run_ostir_one_sequence_new_aSD(self):
        """
        Test run_ostir on one sequence
        """

        print("\n" + "Unit test: run_ostir_one_sequence_new_aSD")

        test_result = ostir.run_ostir("ACUUCUAAUUUAUUCUAUUUAUUCGCGGAUAUGCAUAGGAGUGCUUCGAUGUCAU", start=31,
                                      aSD="ACGTCCCTA")

        # print(test_result)

        expected_result = [
            {
                'name': 'unnamed',
                'start_codon': 'AUG',
                'start_position': 31,
                'expression': 367.8105,
                'RBS_distance_bp': 4,
                'dG_total': 3.4289,
                'dG_rRNA:mRNA': -2.581,
                'dG_mRNA': -7.2,
                'dG_spacing': 0.0039,
                'dG_standby': 0.0,
                'dG_start_codon': -1.194
            },
        ]

        # print(list(test_result))
        # print(list(expected_result))
        self.assertEqual(test_result, expected_result)

class test_unit_constraints(unittest.TestCase):
    def test_unit_constraints(self):
        """
        Tests running with arbitrary constraints
        """
        print("\n" + "Unit test: constraints")
        test_result = ostir.run_ostir("ACUUCUAAUUUAUUCUAUUUAUUCGCGGAUAUGCAUAGGAGUGCUUCGAUGUCAU", start=31,
                                      aSD="ACGTCCCTA", constraints=".....xxxx.")
        expected_result = [{'name': 'unnamed',
                            'start_codon': 'AUG',
                            'start_position': 31,
                            'expression': 886.8033,
                            'RBS_distance_bp': 4,
                            'dG_total': 1.2289,
                            'dG_rRNA:mRNA': -2.581,
                            'dG_mRNA': -5.0,
                            'dG_spacing': 0.0039,
                            'dG_standby': 0.0,
                            'dG_start_codon': -1.194}]

        self.assertEqual(test_result, expected_result)


##############################################################################################
# Integration tests (command line calls)
##############################################################################################

def setUpModule():
    'Handles creation of the output directory and removal of any stale files from a previous test run.'
    # Delete existing directory
    try:
        shutil.rmtree(output_directory_path)
        print(f"Previous 'output' directory removed: {output_directory_path}")
    except OSError as e:
        pass

    # Create fresh output directory
    try:
        os.mkdir(output_directory_path)
        print(f"Created 'output' directory: {output_directory_path}")
    except OSError as e:
        print(f"Failed to create 'output' directory: {output_directory_path}")
        exit(1)


class test_integration_command_line_FASTA_input(unittest.TestCase):
    def test_integration_command_line_FASTA_input(self):
        input_path = os.path.join(THIS_DIR, 'input', 'command_line_FASTA_input.fa')
        output_path = os.path.join(THIS_DIR, 'output', 'command_line_FASTA_input.csv')
        expected_path = os.path.join(THIS_DIR, 'expected', 'command_line_FASTA_input.csv')
        the_command = f'ostir -j 4 -i "{input_path}" -o "{output_path}" -v 0'
        print("\n" + the_command)
        subprocess.call(the_command, shell=True)
        self.assertEqual(True, csv_are_identical(output_path, expected_path))


class test_integration_command_line_string_input(unittest.TestCase):
    def test_integration_command_line_string_input(self):
        output_path = os.path.join(THIS_DIR, 'output', 'command_line_string_input.csv')
        expected_path = os.path.join(THIS_DIR, 'expected', 'command_line_string_input.csv')
        input_sequence = "TTCTAGAAAAAAAATAAGGAGGTAAAATGGCGAGCTCTGAAGACGTTATCAAAGAGTTCATGCGTTTCAAAGTTCGTATG"
        the_command = f'ostir -j 4 -p -i "{input_sequence}" -o "{output_path}" -v 0'
        print("\n" + the_command)
        subprocess.call(the_command, shell=True)
        self.assertEqual(True, csv_are_identical(output_path, expected_path))


class test_integration_command_line_CSV_input(unittest.TestCase):
    def test_integration_command_line_CSV_input(self):
        input_path = os.path.join(THIS_DIR, 'input', 'command_line_CSV_input.csv')
        output_path = os.path.join(THIS_DIR, 'output', 'command_line_CSV_input.csv')
        expected_path = os.path.join(THIS_DIR, 'expected', 'command_line_CSV_input.csv')
        the_command = f'ostir -j 4 -i "{input_path}" -o "{output_path}" -v 0'
        print("\n" + the_command)
        subprocess.call(the_command, shell=True)
        self.assertEqual(True, csv_are_identical(output_path, expected_path))


class test__integration_command_line_CSV_input_alternate_columns_and_defaults(unittest.TestCase):
    def test_integration_command_line_CSV_input(self):
        input_path = os.path.join(THIS_DIR, 'input', 'command_line_CSV_input_alternate_columns_and_defaults.csv')
        output_path = os.path.join(THIS_DIR, 'output', 'command_line_CSV_input_alternate_columns_and_defaults.csv')
        expected_path = os.path.join(THIS_DIR, 'expected', 'command_line_CSV_input_alternate_columns_and_defaults.csv')
        the_command = f'ostir -j 4 -a TCTGAAGAC -p -q -i "{input_path}" -o "{output_path}" -v 0'
        print("\n" + the_command)
        subprocess.call(the_command, shell=True)
        self.assertEqual(True, csv_are_identical(output_path, expected_path))


class test_integration_Salis2009(unittest.TestCase):
    def test_integration_Salis2009(self):
        """
        Test of running Salis2009 sequences
        """

        input_path = os.path.join(THIS_DIR, 'input', 'Salis2009.csv')
        output_path = os.path.join(THIS_DIR, 'output', 'Salis2009.csv')
        expected_path = os.path.join(THIS_DIR, 'expected', 'Salis2009.csv')
        the_command = f'ostir -j 8 -p -i "{input_path}" -o "{output_path}" -v 0'
        print("\n" + the_command)
        subprocess.call(the_command, shell=True)
        self.assertEqual(True, csv_are_identical(output_path, expected_path))


class test_integration_T7_genome(unittest.TestCase):
    def test_integration_T7_genome(self):
        """
        Test of running T7 genome sequence
        """

        input_path = os.path.join(THIS_DIR, 'input', 'T7_genome.fasta')
        output_path = os.path.join(THIS_DIR, 'output', 'T7_genome.csv')
        expected_path = os.path.join(THIS_DIR, 'expected', 'T7_genome.csv')
        the_command = f'ostir -j 8 -i "{input_path}" -o "{output_path}" -v 0'
        print("\n" + the_command)
        subprocess.call(the_command, shell=True)
        self.assertEqual(True, csv_are_identical(output_path, expected_path))

if __name__ == "__main__":
    unittest.main()





# This module contains the rewrite of the ostir factory. It enables running multiple sequences at one.
# This has the result of cleaning up the CLI for the progress bar.


# ------------ Imports ------------
import importlib.util
import concurrent.futures
from typing import Callable
from dataclasses import dataclass

from .ostir_calculations import find_start_codons
from .ostir_worker import ostir_worker
from .data_classes import ostir_task


# Optional dependency
if importlib.util.find_spec("progress") is not None:
    from progress.bar import Bar
else:
    Bar = None

# ------------ Supporting Code ------------


def _ostir_progress_callback(tasks: int) -> Callable:
    # Update the progress bar
    if Bar is not None:
        bar = Bar('Running OSTIR RBS Predictions: ', max=tasks)
    else:
        return None
    max = tasks
    itterations = 0

    def _callback():
        nonlocal itterations
        itterations += 1
        nonlocal max
        if itterations == max:
            bar.next()
            bar.finish()
            return None

        bar.next()
        return None


    return _callback


# -------------- Main Code ------------
def ostir_farm(task_objects: list, threads: int=1, callback: Callable=None, verbosity: int=0) -> dict:
    """ Run OSTIR on multiple sequences in parallel. """

    # Find all start codons in every supplied sequence.
    # Note: returns list of generators
    start_codons = [find_start_codons(task.sequence, start_range=[task.start, task.end]) for task in task_objects]
    id_values = [i for i in range(len(task_objects))]

    # Tie the generators to the task objects
    sets = [(task, start_codon, id_value) for task, start_codon, id_value in zip(task_objects, start_codons, id_values)]

    # Unpack the generators
    arguments = [(task, start_codon, id_value, verbosity) for
                  task, start_codons, id_value in sets for start_codon in start_codons]



    # Tell the callback function how many tasks we have. Expect back another callback function.
    if callback is not None:
        callback = callback(len(arguments))
    elif Bar and verbosity > 0:
        callback = _ostir_progress_callback(len(arguments))
    else:
        callback = lambda: None

    # Create a pool of workers and run OSTIR

    if threads > 1:
        with concurrent.futures.ProcessPoolExecutor(max_workers=threads) as multiprocessor:
            #parallel_output = multiprocessor.map(self._parallel_dG, *parallelizer_arguments)

            futures: list = []
            for task in arguments:
                future_object = multiprocessor.submit(ostir_worker, *task)
                future_object.add_done_callback(lambda x: callback())
                futures.append(future_object)
            parallel_output = [future.result() for future in futures]
    else:
        parallel_output: list = []
        for task in arguments:
            result = ostir_worker(*task)
            parallel_output.append(result)
            callback()

    parallel_output = [x for x in parallel_output if x is not None]

    # Sort out to each ID back the results
    results = {}
    for [id, result] in parallel_output:
        if id not in results:
            results[id] = {'name': result.name,
                                        'results': []}
        results[id]['results'].append(result)

    # Sort the results within each ID
    for item in results.items():
        item[1]['results'].sort(key=lambda x: x.start_position)

    # Return the results
    return parallel_output
    
# Don't forget the end of file newline!



from dataclasses import dataclass

@dataclass
class OSTIRResult():
    """Class for storing the results of an OSTIR calculation"""
    # These are the results that most people care about
    name: str = ""
    start_codon: str = ""
    start_position: int = 0
    expression: float = 0.0
    RBS_distance_bp: int = 0
    dG_total: float = 0.0
    dG_rRNA_mRNA: float = 0.0
    dG_mRNA: float = 0.0
    dG_spacing: float = 0.0
    dG_standby: float = 0.0
    dG_start_codon: float = 0.0

    # These are the results that are not as important as the above
    mRNA_structure: str = ""
    mRNA_rRNA_uncorrected_structure: str = ""
    mRNA_rRNA_corrected_structure: str = ""
    most_5p_mRNA: str = ""
    longest_helix: str = ""
    dG_start_energy: float = 0.0
    helical_loop: str = ""
    bulge_loop: str = ""
    min_bp_prob: float = 0.0
    kinetic_score: float = 0.0

    def __getitem__(self, key):
        if key in self.__annotations__:
            return getattr(self, key)
        raise KeyError(key)

    def results(self):
        """Returns a dictionary of the core results"""
        return {
            'name': self.name,
            'start_codon': self.start_codon,
            'start_position': self.start_position,
            'expression': self.expression,
            'RBS_distance_bp': self.RBS_distance_bp,
            'dG_total': self.dG_total,
            'dG_rRNA:mRNA': self.dG_rRNA_mRNA,
            'dG_mRNA': self.dG_mRNA,
            'dG_spacing': self.dG_spacing,
            'dG_standby': self.dG_standby,
            'dG_start_codon': self.dG_start_codon,
        }

    def addional_results(self):
        """Returns a dictionary additional results"""
        return {
            'mRNA_structure': self.mRNA_structure,
            'mRNA_rRNA_uncorrected_structure': self.mRNA_rRNA_uncorrected_structure,
            'mRNA_rRNA_corrected_structure': self.mRNA_rRNA_corrected_structure,
            'most_5p_mRNA': self.most_5p_mRNA,
            'longest_helix': self.longest_helix,
            'dG_start_energy': self.dG_start_energy,
            'helical_loop': self.helical_loop,
            'bulge_loop': self.bulge_loop,
            'min_bp_prob': self.min_bp_prob,
            'kinetic_score': self.kinetic_score,
        }

@dataclass
class ostir_task:
    sequence: str
    start: int
    end: int
    name: None
    rrna: str='ACCTCCTTA'
    decimals: int=2
    circular: bool=False
    constraints: str=None




#!/usr/bin/env python
"""
This module contains an early rewrite of the OSTIRFactory class. It is not yet
implemented, but it will be used to replace the OSTIRFactory class in the future.

Parrellism is more straightforward to implement in function based code, so this
module turns the class based legacy code into a function based implementation.

Some of the codebase of this file originated from the Salis Lab's RBS calculator which is distributed under GPL3.
See <http://www.gnu.org/licenses/>.
Copyright 2008-2009 is owned by the University of California Regents. All rights reserved.
"""

raise NotImplementedError("This module is not yet implemented")

from dataclasses import dataclass

@dataclass
class OSTIRResult():
    """Class for storing the results of an OSTIR calculation"""
    # These are the results that most people care about
    name: str = ""
    start_codon: str = ""
    start_position: int = 0
    expression: float = 0.0
    RBS_distance_bp: int = 0
    dG_total: float = 0.0
    dG_rRNA_mRNA: float = 0.0
    dG_mRNA: float = 0.0
    dG_spacing: float = 0.0
    dG_standby: float = 0.0
    dG_start_codon: float = 0.0

    # These are the results that are not as important as the above
    mRNA_structure: str = ""
    mRNA_rRNA_uncorrected_structure: str = ""
    mRNA_rRNA_corrected_structure: str = ""
    most_5p_mRNA: str = ""
    longest_helix: str = ""
    dG_start_energy: float = 0.0
    helical_loop: str = ""
    bulge_loop: str = ""
    min_bp_prob: float = 0.0
    kinetic_score: float = 0.0

    def __getitem__(self, key):
        if key in self.__annotations__:
            return getattr(self, key)
        raise KeyError(key)

    def results(self):
        """Returns a dictionary of the core results"""
        return {
            'name': self.name,
            'start_codon': self.start_codon,
            'start_position': self.start_position,
            'expression': self.expression,
            'RBS_distance_bp': self.RBS_distance_bp,
            'dG_total': self.dG_total,
            'dG_rRNA:mRNA': self.dG_rRNA_mRNA,
            'dG_mRNA': self.dG_mRNA,
            'dG_spacing': self.dG_spacing,
            'dG_standby': self.dG_standby,
            'dG_start_codon': self.dG_start_codon,
        }

    def addional_results(self):
        """Returns a dictionary additional results"""
        return {
            'mRNA_structure': self.mRNA_structure,
            'mRNA_rRNA_uncorrected_structure': self.mRNA_rRNA_uncorrected_structure,
            'mRNA_rRNA_corrected_structure': self.mRNA_rRNA_corrected_structure,
            'most_5p_mRNA': self.most_5p_mRNA,
            'longest_helix': self.longest_helix,
            'dG_start_energy': self.dG_start_energy,
            'helical_loop': self.helical_loop,
            'bulge_loop': self.bulge_loop,
            'min_bp_prob': self.min_bp_prob,
            'kinetic_score': self.kinetic_score,
        }


class OstirManager():
    """This is the class that accepts sequences, prepares parallelization, and sends parts out for computation.
    Everything here is managed in python, while the actual computation is sent off to cython"""
    def __init__(self):
        self.start_range_1 = None
        self.start_codons = None

    @staticmethod
    def _job() -> OSTIRResult:
        
        pass

    def run(self) -> list:
        """Runs the OSTIR calculation on the submitted sequences"""
        
        
    def find_start_codons(self, sequence):
        """Finds all start codons in an mRNA sequence. Creates a list."""

        self.start_position_list = []

        seq_len = len(sequence)

        #Switch to zero-indexed positions happens here
        end_0 = min(self.start_range_1[1]-1, seq_len - 2)
        begin_0 = min(self.start_range_1[0]-1, end_0)

        for i in range(begin_0, end_0 + 1):
            codon = sequence[i:i + 3]
            if codon.upper() in self.start_codons:
                self.start_position_list.append(i)
                yield (i, codon)
            else:
                pass
    
    def submit_sequence_str(self):
        pass

    def submit_sequence_file(self):
        pass



"""
Shortcuts for common tasks
"""

import sys
import csv
from itertools import groupby
from .data_classes import ostir_task
from .ostir_farm import ostir_farm

# ----- Process CSV's ----

def from_csv(csv_file,
             threads=1,
             verbosity=0,
             **kwargs):
    """Parses a given CSV and runs OSTIR on it. Returns a dictionary of results. Additional kwargs are used as defaults."""
    asd = kwargs.get('asd', 'ACCTCCTTA') or 'ACCTCCTTA'
    start = kwargs.get('start', 0) or 0
    end = kwargs.get('end', None)  # None means the end of the sequence
    circular = kwargs.get('circular', False)
    constraints = kwargs.get('constraints', None)
    name = kwargs.get('name', None) # Prefex for unnamed sequences

    tasks = []

    with open(csv_file, 'r', encoding='UTF-8-sig') as f:
        reader = csv.DictReader(BlankCommentCSVFile(f))
        on_seq_index = 0 #used for giving names if none provided in input file
        for row in reader:
            on_seq_index = on_seq_index+1

            #lowercase the keys
            row = dict((k.lower(), v) for k, v in row.items())

            # for debugging
            #print(row)

            ## Allow 'sequence' or 'seq'
            if 'seq' in row.keys():
                sequence = row['seq']
            elif 'sequence' in row.keys():
                sequence = row['sequence']
            else:
                error_string = f"Required column 'sequence' or 'seq' not found for CSV file row: {row}"
                print(error_string, file=sys.stderr)
                sys.exit(1)

            sequence = sequence.replace(' ', '') #remove spaces

            # Assign a name if one is not given from name/id columns
            # If empty assign one based on the index
            name = row.get('name')
            if not name:
                name = row.get('id') or "sequence_" + str(on_seq_index)

            aSD = row.get('anti-shine-dalgarno', asd) or asd #remember, keys lowercased here

            end_loc_1 = row.get('end', end) or end or row.get('start', start) or len(sequence)
            start_loc_1 = row.get('start', start) or start or 1
            circular = row.get('circular', circular) or circular
            constraints = row.get('constraints', constraints) or constraints

            if not circular:
                pass
            elif circular.lower() in ['true', 't', 'yes', 'y', '1']:
                circular = True
            elif circular.lower() in ['false', 'f', 'no', 'n', '0']:
                circular = False
            else:
                print(f"Invalid value for circular: {circular}", file=sys.stderr)
                sys.exit(1)

            # Make an ostir_task from the collected data
            task = ostir_task(sequence,
                              name=name,
                              rrna=aSD,
                              start=int(start_loc_1),
                              end=int(end_loc_1),
                              circular=circular,
                              constraints=constraints)
            tasks.append(task)

        # If any of these are not defined or are empty, use command line values as the defaults

    results = ostir_farm(tasks, threads=threads, verbosity=verbosity)
    # group results by the first value in the list and sort by start position
    results = {k: list(v) for k, v in groupby(results, lambda x: x[0])}
    for key, value in results.items():
        results[key] = sorted([data for _, data in value], key=lambda x: x.start_position)
        

    # Insure the dictionary is sorted by key
    results = {k: results[k] for k in sorted(results)}

    return results

# Used to get rid of blank and comment lines on file input into csv.reader
class BlankCommentCSVFile:
    def __init__(self, fp):
        self.fp = fp

    def __iter__(self):
        return self

    def __next__(self):
        line = self.fp.__next__()
        if not line.strip() or line[0] == "#":
            return self.__next__()
        return line

# ----- Process FASTA's ----

def from_fasta():
    pass

# ----- Process strings ----

def from_string():
    pass

# ----- Process GenBank's ----

def from_genbank():
    pass



# ---- Simplify outputs ----
def simplify_results(results):
    """Takes a dictionary of results and returns a simplified dictionary of results"""
    simplified_results = {}
    for key, value in results.items():
        simplified_results[key] = value.simplify()
    return simplified_results




#!/usr/bin/env python
"""
This module contains the OSTIRFactory class, which is the main script used to call the ostir calculation functions
and compile a result. There should be no need to interact with this module directly except in advanced use cases.

Some of the codebase of this file originated from the Salis Lab's RBS calculator which is distributed under GPL3.
See <http://www.gnu.org/licenses/>.
Copyright 2008-2009 is owned by the University of California Regents. All rights reserved.
"""


import re
import math
import os
import concurrent.futures
import importlib.util
from dataclasses import dataclass
from .ostir_calculations import calc_longest_loop_bulge, calc_longest_helix, calc_dG_standby_site, find_start_codons, calc_dG_mRNA, calc_dG_mRNA_rRNA, calc_expression_level

# Optional dependency
if importlib.util.find_spec("progress") is not None:
    from progress.bar import Bar
else:
    Bar = None


class CalcError(Exception):
    """Base class for exceptions in this module."""

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)

@dataclass
class OSTIRResult():
    """Class for storing the results of an OSTIR calculation"""
    # These are the results that most people care about
    name: str = ""
    start_codon: str = ""
    start_position: int = 0
    expression: float = 0.0
    RBS_distance_bp: int = 0
    dG_total: float = 0.0
    dG_rRNA_mRNA: float = 0.0
    dG_mRNA: float = 0.0
    dG_spacing: float = 0.0
    dG_standby: float = 0.0
    dG_start_codon: float = 0.0

    # These are the results that are not as important as the above
    mRNA_structure: str = ""
    mRNA_rRNA_uncorrected_structure: str = ""
    mRNA_rRNA_corrected_structure: str = ""
    longest_helix: str = ""
    dG_start_energy: float = 0.0
    helical_loop: str = ""
    bulge_loop: str = ""
    min_bp_prob: float = 0.0
    kinetic_score: float = 0.0

    def __getitem__(self, key):
        return getattr(self, key)

    def results(self):
        """Returns a dictionary of the core results"""
        return {
            'name': self.name,
            'start_codon': self.start_codon,
            'start_position': self.start_position,
            'expression': self.expression,
            'RBS_distance_bp': self.RBS_distance_bp,
            'dG_total': self.dG_total,
            'dG_rRNA:mRNA': self.dG_rRNA_mRNA,
            'dG_mRNA': self.dG_mRNA,
            'dG_spacing': self.dG_spacing,
            'dG_standby': self.dG_standby,
            'dG_start_codon': self.dG_start_codon,
        }

    def addional_results(self):
        """Returns a dictionary additional results"""
        return {
            'mRNA_structure': self.mRNA_structure,
            'mRNA_rRNA_uncorrected_structure': self.mRNA_rRNA_uncorrected_structure,
            'mRNA_rRNA_corrected_structure': self.mRNA_rRNA_corrected_structure,
            'longest_helix': self.longest_helix,
            'dG_start_energy': self.dG_start_energy,
            'helical_loop': self.helical_loop,
            'bulge_loop': self.bulge_loop,
            'min_bp_prob': self.min_bp_prob,
            'kinetic_score': self.kinetic_score,
        }


class OSTIRFactory:
    """Class for calculating the rate of translation initiation of a given mRNA sequence"""
    def __init__(self, mRNA, start_range, rRNA, constraints, verbosity=0, decimal_places=4, circular=False, name="unnamed"):
        """
        Initializes the RBS Calculator class with the mRNA sequence and the range of start codon positions considered.
        start_range is a pair of 1-indexed positions
        """

        # Sets deaults
        # From OSTIR calibration using Salis2009 data. See calibration directory for procedure
        self.Beta = 0.40002512
        self.RT_eff = 1/self.Beta
        self.logK = 7.279194329
        self.K = math.exp(self.logK)

        # Global parameters -- constants
        self.infinity = 1e12  # For all practical purposes, here.
        self.RNA_model = 'rna2004'
        self.start_codon_energies = {"ATG": -1.194, "AUG": -1.194, "GTG": -0.0748, "GUG": -0.0748, "TTG": -0.0435,
                                        "UUG": -0.0435, "CTG": -0.03406, "CUG": -0.03406}  # hybridization to CAT
        self.auto_dangles = True
        self.dangles_default = "all"
        self.temp = 37.0
        self.optimal_spacing = 5  # aligned spacing

        # From OSTIR calibration using Salis2009 data. See calibration directory for procedure
        self.dG_spacing_constant_push = [17.20965071, 3.46341492, 1.790848365, 3.0]
        self.dG_spacing_constant_pull = [0.06422042, 0.275640836, 0.0]
        self.cutoff = 35  # number of nt +- start codon considering for folding
        self.standby_site_length = 4  # Number of nt before SD sequence that must be unpaired for ribosome binding
        self.energy_cutoff = 3.0
        self.start_codons = ["ATG", # substituted U for T in actual calcs. Ignores CTG/CUG
                             "AUG",
                             "GTG",
                             "GUG",
                             "TTG",
                             "UUG"]
        self.footprint = 1000
        """Footprint of the 30S complex that prevents formation of secondary structures
        downstream of the start codon. Here, we assume that the entire post-start RNA
        sequence does not form secondary structures once the 30S complex has bound.
        """
        self.circular = circular

        # NuPACK.__init__(self,sequences,self.RNA_model)
        exp = re.compile('[ATGCU._]', re.IGNORECASE)
        if exp.match(mRNA) is None:
            raise ValueError(f"Invalid letters found in sequence {mRNA}. Only ATGCU accepted.")
        mRNA = mRNA.replace('.', '')
        mRNA = mRNA.replace('_', '')

        if start_range[0] < 1:
            start_range[0] = 1
        if start_range[1] > len(mRNA):
            start_range[1] = len(mRNA)


        self.install_location = os.path.dirname(os.path.realpath(__file__))
        self.name = name
        self.mRNA_input = mRNA.upper()
        self.rRNA = rRNA
        self.constraints = constraints
        self.rRNA_len = len(self.rRNA)
        self.mRNA_len = len(self.mRNA_input)
        self.total_sequence_length = len(mRNA) + len(self.rRNA)
        self.run = 0
        self.start_range = start_range
        self.verbosity = verbosity
        self.threads = 1
        self.decimal_places = decimal_places
        self.results = []


    def calc_dG(self):
        """Calculates each dG term in the free energy model and sums them together to create dG_total"""

        mRNA_length = len(self.mRNA_input)
        if self.circular:
            if mRNA_length > 200:
                self.mRNA_input = self.mRNA_input + self.mRNA_input[:200]
            else:
                self.mRNA_input = self.mRNA_input + self.mRNA_input[:mRNA_length]

        # Set up progress bar for optional dependency

        arguments = []

        parallelizer_arguments = [[], [], []]
        for i, (start_pos, codon) in enumerate(find_start_codons(self.mRNA_input[:mRNA_length], self.start_range)):
            parallelizer_arguments[0].append(self)
            parallelizer_arguments[1].append(start_pos)
            parallelizer_arguments[2].append(codon)
            arguments.append([self, start_pos, codon])
        if Bar and self.verbosity > 0:
            progress_bar = Bar('Running OSTIR RBS Predictions', max=i+1)
        else:
            progress_bar = None


        if self.threads > 1:
            with concurrent.futures.ProcessPoolExecutor(max_workers=self.threads) as multiprocessor:
                #parallel_output = multiprocessor.map(self._parallel_dG, *parallelizer_arguments)

                futures = []
                for task in arguments:
                    future_object = multiprocessor.submit(self._parallel_dG, *task)
                    if progress_bar:
                        future_object.add_done_callback(lambda x: progress_bar.next())
                    futures.append(future_object)
                parallel_output = [future.result() for future in futures]
        else:
            parallel_output = []
            for task in arguments:
                result = self._parallel_dG(*task)
                parallel_output.append(result)
                if progress_bar:
                    progress_bar.next()
        if progress_bar:
            progress_bar.finish()

        parallel_output = [x for x in parallel_output if x is not None]
        self.results = parallel_output

        self.run = 1


    @staticmethod
    def _parallel_dG(factory_obj, start_pos, codon):
        try:

            # print "Top of calc_dG here"

            # Set dangles based on length between 5' end of mRNA and start codon
            if factory_obj.auto_dangles:
                cutoff = int(factory_obj.cutoff)
                if start_pos > cutoff:
                    dangles = "none"

                else:
                    dangles = "all"

            else:
                dangles = factory_obj.dangles_default
                # print "Auto Dangles set to ", self.dangles

            # Start codon energy
            dG_start_codon = factory_obj.start_codon_energies[codon]

            # Energy of mRNA folding. This also gets us the kinetic score
            dG_mRNA, mRNA_structure, kinetic_score, min_bp_prob = calc_dG_mRNA(factory_obj.mRNA_input, start_pos, dangles, factory_obj.constraints)

            # Energy of mRNA:rRNA hybridization & folding
            dg_mRNA_rRNA_output = calc_dG_mRNA_rRNA(factory_obj.mRNA_input, factory_obj.rRNA, start_pos, dangles, factory_obj.constraints)
            dG_mRNA_rRNA_withspacing = dg_mRNA_rRNA_output[0]
            mRNA_rRNA_structure = dg_mRNA_rRNA_output[1]
            spacing_value = dg_mRNA_rRNA_output[2]
            if not dG_mRNA_rRNA_withspacing:
                return None

            dG_mRNA_rRNA_withspacing -= 2.481  # Modifying hybridization penalty to match NUPACK

            dG_mRNA_rRNA_nospacing = mRNA_rRNA_structure["dG_mRNA_rRNA"]
            dG_mRNA_rRNA_nospacing -= 2.481  # Modifying hybridization penalty to match NUPACK

            # Standby site correction:
            dG_standby_site, corrected_structure = calc_dG_standby_site(mRNA_rRNA_structure,
                                                                                    dangles,
                                                                                    factory_obj.standby_site_length,
                                                                                    factory_obj.constraints,
                                                                                    factory_obj.rRNA)

            # Total energy is mRNA:rRNA + start - rRNA - mRNA - standby_site
            dG_total = dG_mRNA_rRNA_withspacing + dG_start_codon - dG_mRNA - dG_standby_site

            # Calculate dG to open SD sequence
            # ddG_SD_open = self.calc_dG_SDopen(mRNA_structure, mRNA_rRNA_structure)

            loop_result = calc_longest_loop_bulge(structure=mRNA_structure)
            helical_loop_list, bulge_loop_list = loop_result[0], loop_result[1]

            parallel_output = OSTIRResult(
                name= factory_obj.name,
                start_codon=codon,
                start_position=start_pos+1,
                expression=round(calc_expression_level(dG_total),
                                 factory_obj.decimal_places),
                RBS_distance_bp=int(spacing_value),
                dG_total=round(float(dG_total), factory_obj.decimal_places),
                dG_rRNA_mRNA=round(float(dG_mRNA_rRNA_nospacing), factory_obj.decimal_places),
                dG_mRNA=round(float(dG_mRNA), factory_obj.decimal_places),
                dG_spacing=round(float(mRNA_rRNA_structure["dG_spacing"]), factory_obj.decimal_places),
                dG_standby=round(float(dG_standby_site), factory_obj.decimal_places),
                dG_start_codon=round(float(dG_start_codon), factory_obj.decimal_places),

                mRNA_structure=mRNA_structure,
                mRNA_rRNA_uncorrected_structure=mRNA_rRNA_structure,
                mRNA_rRNA_corrected_structure=corrected_structure,
                longest_helix=calc_longest_helix(mRNA_structure),
                dG_start_energy=dG_start_codon,
                helical_loop=helical_loop_list,
                min_bp_prob=min_bp_prob,
                bulge_loop=bulge_loop_list,
                kinetic_score=kinetic_score
            )

        except ValueError as msg:
            if "leaderless start codon" not in str(msg):
                raise ValueError(msg) from msg
            print(msg)
            parallel_output = None

        return parallel_output


# ----------------------------------------------------------------------------------------------------------
# End RBS_Calculator class
# ----------------------------------------------------------------------------------------------------------



#!/usr/bin/env python
'''This module is the main module for the OSTIR package. It contains the run_ostir function for
high level execution and the CLI interface. To use, run "ostir" in the command line or import
ostir and run run_ostir.

OSTIR is distributed under GPL3. See <http://www.gnu.org/licenses/>.'''


import argparse
import os
import sys
import subprocess
import csv
import re
import importlib.util
from shutil import which
from warnings import warn
from pathlib import Path
from .shortcuts import from_csv


try:
    from ostir.ostir_factory import OSTIRFactory
except ModuleNotFoundError:
    from .ostir_factory import OSTIRFactory

# Optional dependency
if importlib.util.find_spec("rich") is not None:
    from rich import print as rprint
else:
    rprint = None

OSTIR_VERSION = '1.1.1'
OLDEST_VIENNA = '2.4.18'

# The E. coli sequence
Ecoli_anti_Shine_Dalgarno = 'ACCTCCTTA'

def run_ostir(in_seq,
              start=None,
              end=None,
              name=None,
              aSD=None,
              threads=1,
              decimal_places=4,
              circular=False,
              constraints=None,
              verbosity=0):
    '''Takes an RNA with optional parameters and returns binding energies.
        Keyword arguments:
        seq -- Sequence to calculate binding energies for
        outfile -- Filepath for output csv
        start -- First base to start considering start cofdons. 1-indexed. Defaults to first base
        end -- Last base to start considering start codons. 1-indexed.
                     If start_loc_1 is provided, defaults to start_loc_1
                     Otherwise, defaults to end of sequence
        name -- Returns itself, useful for tagging things for downstream processing
        aSD -- Defines anti-Shine-Dalgarno sequence: the 9 bp at the 3' end of the 16S rRNA. Defaults to the E. coli sequence.
        threads -- Defines parallel processing workers, roughly equivalent to multithreading cores
        decimal_places -- Precision of numerical output (number of places to the right of the decimal)
        constraints -- Folding constraints passed to vienna starting from the cuttoff window (35bp upstream of start codon)
        verbosity -- Controls verbosity of output. 0 is silent, 1 is normal, 2 is verbose
    '''
    #rename to more descriptive vars
    in_start_loc_1 = start
    in_end_loc_1 = end

    # Set up a default name and sd if not provided
    if name is None:
        name = 'unnamed'

    if aSD is None:
        aSD = Ecoli_anti_Shine_Dalgarno

    # Do some checks on the inputs

    # Nucleotide character check
    nucleotides=re.compile('[^ATCGUatcgu]')
    if nucleotides.search(aSD) is not None:
        error_message = f"ERROR: anti-Shine-Dalgarno sequence provided ({aSD}) contains non-nucleotide characters.\n<<<Sequence ({name}) will be skipped.>>>"
        if not rprint:
            print(error_message, file=sys.stderr)
        else:
            rprint(error_message, style="bold red")
        return []

    # Length check
    if len(aSD) != 9:
        error_message = f"ERROR: anti-Shine-Dalgarno sequence provided ({aSD}) is not 9 bases.\n<<<Sequence ({name}) will be skipped.>>>"
        if not rprint:
            print(error_message, file=sys.stderr)
        else:
            rprint(error_message, style="bold red")
        return []

    #Upper case and convert to RNA
    aSD = aSD.upper()
    aSD = aSD.replace("T","U")

    #Clean spaces from sequence... could clean other characters too
    seq = in_seq.replace(" ", "")

    # Nucleotide character check
    if nucleotides.search(seq) is not None:
        error_message = f"ERROR: Input sequence contains non-nucleotide characters.\n<<<Sequence ({name}) will be skipped.>>>"
        if not rprint:
            print(error_message, file=sys.stderr)
        else:
            rprint(error_message, style="bold red")
        return []


    # Start <= end check
    if in_start_loc_1 is not None and in_end_loc_1 is not None and in_end_loc_1<in_start_loc_1:
        error_message = f"ERROR: Start location ({in_start_loc_1}) is not less than end location ({in_end_loc_1}).\n<<<Sequence ({name}) will be skipped.>>>"
        if not rprint:
            print(error_message, file=sys.stderr)
        else:
            rprint(error_message, style="bold red")
        return []

    # Set up start and end locations to search
    start_loc_1 = in_start_loc_1
    if start_loc_1 is None:
        start_loc_1 = 1
    start_loc_1 = int(start_loc_1)

    end_loc_1 = in_end_loc_1
    if end_loc_1 is None:
        if in_start_loc_1 is None:
            end_loc_1 = len(seq)
        else:
            end_loc_1 = start_loc_1
    end_loc_1 = int(end_loc_1)

    start_range_1 = [start_loc_1, end_loc_1]

    calcObj = OSTIRFactory(seq, start_range_1, aSD, constraints, circular=circular, verbosity=verbosity)
    calcObj.threads = threads
    calcObj.decimal_places = decimal_places
    calcObj.name = name
    calcObj.calc_dG()

    output_data_list = [result.results() for result in calcObj.results]

    output_data_list = sorted(output_data_list, key=lambda x: x['start_position'])

    del calcObj  # Send OSTIRFactory to garbage collection, clears temp files

    return output_data_list

def parse_fasta(filepath):
    '''Takes a filepath to a fasta formatted file and returns a list of [header, sequence].'''
    sequences = []
    current_seq_name = None
    current_seq = ""
    with open(filepath, 'r', encoding='utf8') as infile:
        for line in infile:
            linestr = str(line)
            if linestr[0] == '>':
                if current_seq_name:
                    sequences.append([current_seq_name, current_seq])
                current_seq_name = linestr[1:].rstrip()
                current_seq = str()
                continue
            else:
                current_seq += linestr.rstrip()
    if current_seq_name:
        sequences.append([current_seq_name, current_seq])
    return sequences

def _print_output(outdict):
    '''Processes command line / csv input for pretty command line output'''
    sorted_predictions = {}
    keys = []
    for prediction in outdict:
        if prediction['name'] in sorted_predictions:
            sorted_predictions[prediction['name']].append(prediction)
        else:
            sorted_predictions[prediction['name']] = [prediction]
            out_names = []
            out_names.append(prediction['name'])
            if 'sequence' in prediction:
                out_names.append(prediction['sequence'])
            else:
                out_names.append(None)
            if 'anti-Shine-Dalgarno' in prediction:
                out_names.append(prediction['anti-Shine-Dalgarno'])
            else:
                out_names.append(None)
            keys.append(out_names)

    if not keys:
        if not rprint:
            print('No binding sites were identified.')
        else:
            rprint('[bold red]No binding sites were identified.[/bold red]')
        sys.exit(0)
    output_items = ['start_codon', 'start_position', 'expression', 'RBS_distance_bp', 'dG_total', 'dG_rRNA:mRNA', 'dG_mRNA', 'dG_spacing', 'dG_standby', 'dG_start_codon']
    if not rprint:
        row_format = "{:>16}" * (len(output_items))
        print('_________________________________________________')
        for rna in keys:
            print(f'\nSample: {rna[0]}')
            if rna[1]:
                print(f'Tested Sequence: {rna[1]}')
            if rna[2]:
                print(f'Sequence RNA: {rna[2]}')
            print(row_format.format(*output_items))
            for start in sorted_predictions[rna[0]]:
                output_data = [start[key] for key in output_items]
                for i, data_point in enumerate(output_data):
                    if isinstance(data_point, float):
                        data_point = format(data_point, '.4f')
                        output_data[i] = data_point
                print(row_format.format(*output_data))
            print('_________________________________________________')
    else:
        from rich.table import Table
        from rich.console import Console
        console = Console()
        for rna in keys:
            print(f'\nSample: {rna[0]}')
            if rna[1]:
                print(f'Tested Sequence: {rna[1]}')
            if rna[2]:
                print(f'Sequence RNA: {rna[2]}')
            table = Table()
            for item in output_items:
                table.add_column(item)
            for start in sorted_predictions[rna[0]]:
                output_data = [start[key] for key in output_items]
                for i, data_point in enumerate(output_data):
                    if isinstance(data_point, float):
                        data_point = format(data_point, '.4f')
                        output_data[i] = data_point
                table.add_row(str(start['start_codon']),
                              str(start['start_position']),
                                str(start['expression']),
                                str(start['RBS_distance_bp']),
                                str(start['dG_total']),
                                str(start['dG_rRNA:mRNA']),
                                str(start['dG_mRNA']),
                                str(start['dG_spacing']),
                                str(start['dG_standby']),
                                str(start['dG_start_codon']))

            console.print(table)
            
def save_to_csv(column_names, outdict, outfile):
    with open(outfile, 'w', encoding='utf8') as output_file:
        dict_writer = csv.DictWriter(output_file, column_names, lineterminator="\n")
        dict_writer.writeheader()
        dict_writer.writerows(outdict)

# Used to get rid of blank and comment lines on file input into csv.reader
class BlankCommentCSVFile:
    def __init__(self, fp):
        self.fp = fp

    def __iter__(self):
        return self

    def __next__(self):
        line = self.fp.__next__()
        if not line.strip() or line[0] == "#":
            return self.__next__()
        return line

def main():
    '''Main OSTIR function ran when called as an executable. See ostir -h.'''
    parser = argparse.ArgumentParser(description=f'OSTIR (Open Source Translation Initiation Rates) version {OSTIR_VERSION}')

    parser.add_argument(
        '-i', '--input',
        action='store',
        metavar='str/filepath',
        dest='i',
        required=False,
        type=str,
        help="Input filename (FASTA/CSV) or DNA/RNA sequence. For CSV input files, there must be a 'seq' or 'sequence' column. Other columns will override any options provided at the command line if they are present: 'name/id', 'start', 'end', 'anti-Shine-Dalgarno'",
    )

    parser.add_argument(
        '-o', '--output',
        action='store',
        metavar= 'filepath',
        dest='o',
        required=False,
        type=str,
        help="Output file path. If not provided, results will output to the console",
    )

    parser.add_argument(
        '-v', '--versity',
        action='store',
        metavar= 'int',
        dest='v',
        required=False,
        default="1",
        help="Sets the verbosity level. Default 0 is quiet, 1 is normal, 2 is verbose",
    )

    parser.add_argument(
        '-s', '--start',
        action='store',
        metavar='int',
        dest='s',
        required=False,
        type=int,
        help="Most 5' nucleotide position to consider a start codon beginning (1-indexed)",
    )

    parser.add_argument(
        '-e', '--end',
        action='store',
        metavar='int',
        dest='e',
        required=False,
        type=int,
        help="Most 3' nucleotide position to consider a start codon beginning (1-indexed)",
    )

    parser.add_argument(
        '-a', '--anti-Shine-Dalgarno',
        action='store',
        metavar='str',
        dest='a',
        required=False,
        type=str,
        help=f"anti-Shine-Dalgarno sequence: the 9 bases located at the 3' end of 16S rRNA. May be provided as DNA or RNA. Defaults to that of E. coli ({Ecoli_anti_Shine_Dalgarno}).",
    )

    parser.add_argument(
        '-p', '--print-sequence',
        action='store_true',
        dest='p',
        required=False,
        help="Include the input mRNA sequence in output CSV files",
    )

    parser.add_argument(
        '-c', '--circular',
        action='store_true',
        dest='c',
        required=False,
        help="Flag the input as circular",
    )

    parser.add_argument(
        '-q', '--print-anti-Shine-Dalgarno',
        action='store_true',
        dest='q',
        required=False,
        help="Include the anti-Shine-Dalgarno sequence in output CSV files",
    )

    parser.add_argument(
        '-j', '--threads',
        action='store',
        metavar='int',
        dest='j',
        required=False,
        type=int,
        help="Number of threads for multiprocessing",
    )

    parser.add_argument(
        '-t', '--type',
        action='store',
        metavar='[string|csv|fasta]',
        dest='t',
        required=False,
        type=str,
        help="Input type (overrides autodetection)",
    )

    parser.add_argument(
        '--version',
        action='store_true',
        dest='version',
        required=False,
        help="Print version and quit.",
    )

    options = parser.parse_args()

    if options.version:
        print_string = f'OSTIR version {OSTIR_VERSION}'
        if not rprint:
            print(print_string, file=sys.stdout)
        else:
            rprint(print_string)
        sys.exit(0)

    if not options.i:
        print_string = "Input (-i) required."
        if not rprint:
            print(print_string, file=sys.stderr)
        else:
            rprint(print_string, file=sys.stderr)
        parser.print_help()
        sys.exit(1)

    cmd_kwargs = dict()
    cmd_kwargs['seq'] = options.i
    if options.o:
        outfile = options.o
    else:
        outfile = None
        cmd_kwargs['print_out'] = False

    if options.j:
        threads = options.j
        cmd_kwargs['threads'] = threads
    else:
        threads = 1
    if options.s:
        cmd_kwargs['start'] = options.s
    if options.e:
        cmd_kwargs['end'] = options.e
    if options.a:
        cmd_kwargs['aSD'] = options.a
    if options.p:
        cmd_kwargs['print_mRNA_sequence'] = options.p
    if options.q:
        cmd_kwargs['print_aSD_sequence'] = options.q
    if options.c:
        cmd_kwargs['circular'] = options.c
    if options.v:
        if options.v not in ["0", "1", "2"]:
            print_string = "Verbosity (-v) must be 0, 1, or 2."
            print(options.v)
            if not rprint:
                print(print_string, file=sys.stderr)
            else:
                rprint(print_string, file=sys.stderr)
            parser.print_help()
            sys.exit(1)
        cmd_kwargs['verbosity'] = int(options.v)
        verbosity = cmd_kwargs['verbosity']

    # Check if viennaRNA is installed
    dependencies = [which('RNAfold') is not None,
                    which('RNAsubopt') is not None,
                    which('RNAeval') is not None]

    if False in dependencies:
        raise EnvironmentError('ViennaRNA is not properly installed or in PATH')

    vienna_version = subprocess.check_output(['RNAfold', '--version'])
    vienna_version = str(vienna_version.strip()).replace("'", "").split(' ')[1]
    if verbosity > 0:
        print_string = f'Running OSTIR version {OSTIR_VERSION} (with Vienna version: {vienna_version})'
        if not rprint:
            print(print_string, file=sys.stdout)
        else:
            rprint(print_string)

    # Check if the viennaRNA version is recent enough
    vienna_version_split = vienna_version.split('.')
    global OLDEST_VIENNA
    OLDEST_VIENNA_split = OLDEST_VIENNA.split('.')
    warning_string = f'The installed version of ViennaRNA ({vienna_version}) is older than what is supported ({OLDEST_VIENNA}).'
    for i, split_digit in enumerate(vienna_version_split):
        if split_digit > OLDEST_VIENNA_split[i]:
            warn(f'The installed version of ViennaRNA ({vienna_version}) is new than what is last validated ({OLDEST_VIENNA}).')
            break
        if split_digit < OLDEST_VIENNA_split[i]:
            raise EnvironmentError(warning_string)
    # Output data: RNA, Codon, position, dg_total, dg rRNA:mRNA, dg mRNA, dG Spacing, dg Standby, Kinetic Score

    # Determine input type
    input_type = None
    specified_file_exists = False
    valid_string_check = re.compile('[ATGCU.-]', re.IGNORECASE)

    if options.t:  # Manual override
        if options.t == 'fasta':
            input_type = 'fasta'
        elif options.t == 'csv':
            input_type = 'csv'
        elif options.t == 'string':
            input_type = 'string'
        else:
            error_string = f'Unsupported file type {options.t}.'
            if not rprint:
                print(error_string, file=sys.stderr)
            else:
                rprint(error_string, file=sys.stderr)
            sys.exit(1)
    elif os.path.isfile(options.i) and not input_type:  # Get input type from file name
        filepath_test = Path(cmd_kwargs['seq']).suffix
        if filepath_test in ['.fasta', '.fa', '.fna']:
            input_type = 'fasta'
        elif filepath_test == '.csv':
            input_type = 'csv'
        else:
            with open(cmd_kwargs['seq'], 'r', encoding='utf8') as in_file:   # Try to determine input type from file contents
                specified_file_exists = True
                first_line = in_file.readline()
                if first_line[0] == '>':
                    input_type = 'fasta'
                elif 'seq' in first_line[0]:
                    input_type = 'csv'

    elif valid_string_check.match(cmd_kwargs['seq']):  # Check to see if input is a valid sequence
        input_type = 'string'

    if input_type is None:   # Error out
        if specified_file_exists:
            error_string = f'Unable to identify the type of file specified as input (-i). Please define it using "-t".'
            if not rprint:
                print(error_string, file=sys.stderr)
            else:
                rprint(error_string, file=sys.stderr)
        else:
            error_string = f'Unable to identify the type of input (-i). Please define it using "-t".'
            if not rprint:
                print(error_string, file=sys.stderr)
            else:
                rprint(error_string, file=sys.stderr)
        sys.exit(1)


    # Run OSTIR
    results = []
    ## String input ##############################################################
    if input_type == 'string':
        if verbosity > 0:
            print_string = 'Reading input sequence from command line'
            if not rprint:
                print(print_string, file=sys.stdout)
            else:
                rprint(print_string)

        sequence = cmd_kwargs.get('seq')
        name = None
        start_loc_1 = cmd_kwargs.get('start')
        end_loc_1 = cmd_kwargs.get('end')
        aSD = cmd_kwargs.get('aSD')
        verbosity = cmd_kwargs.get('verbosity')
        circular = cmd_kwargs.get('circular')

        output_dict_list = run_ostir(sequence,
                                     start=start_loc_1,
                                     end=end_loc_1,
                                     name=name,
                                     aSD=aSD,
                                     threads=threads,
                                     circular=circular,
                                     verbosity=verbosity)

        for output_dict in output_dict_list:
            if cmd_kwargs.get('print_mRNA_sequence'):
                output_dict['sequence'] = sequence
            if cmd_kwargs.get('print_aSD_sequence'):
                output_dict['anti-Shine-Dalgarno'] = aSD

        results.extend(output_dict_list)

    ## FASTA input ##############################################################
    elif input_type == 'fasta':
        input_file = cmd_kwargs['seq']
        if verbosity > 0:
            print_string = f'Reading FASTA file {input_file}'
            if not rprint:
                print(print_string, file=sys.stdout)
            else:
                rprint(print_string)
        sequence_entries = parse_fasta(input_file)
        for sequence_entry in sequence_entries:
            sequence = sequence_entry[1]
            name = sequence_entry[0]
            start_loc_1 = cmd_kwargs.get('start')
            end_loc_1 = cmd_kwargs.get('end')
            aSD = cmd_kwargs.get('aSD')
            circular = cmd_kwargs.get('circular')
            verbosity = cmd_kwargs.get('verbosity')

            output_dict_list = run_ostir(sequence,
                                         start=start_loc_1,
                                         end=end_loc_1,
                                         name=name,
                                         aSD=aSD,
                                         threads=threads,
                                         circular=circular,
                                         verbosity=verbosity)

            for output_dict in output_dict_list:
                if cmd_kwargs.get('print_mRNA_sequence'):
                    output_dict['sequence'] = sequence
                if cmd_kwargs.get('print_aSD_sequence'):
                    output_dict['anti-Shine-Dalgarno'] = aSD

            results.extend(output_dict_list)


    ## CSV input ##############################################################
    elif input_type == 'csv':
        input_file = cmd_kwargs['seq']
        if verbosity > 0:
            print_string = f'Reading CSV file {input_file}'
            if not rprint:
                print(print_string, file=sys.stdout)
            else:
                rprint(print_string)

        result = from_csv(csv_file=input_file,
                        threads=cmd_kwargs.get('threads'),
                        asd=cmd_kwargs.get('aSD'),
                        start=cmd_kwargs.get('start'),
                        end=cmd_kwargs.get('end'),
                        verbosity=cmd_kwargs.get('verbosity'),
                        circular=cmd_kwargs.get('circular'),)

        # We also need the metadata from the CSV file
        run_metadata = []
        with open(input_file, 'r', encoding='UTF-8-sig') as csv_file:
            reader = csv.DictReader(BlankCommentCSVFile(csv_file))
            for row in reader:
                run_metadata.append(row)

        results = []
        for key, finding in result.items():
            result_set = []
            for value in finding:
                value_dict = value.results()
                if cmd_kwargs.get('print_mRNA_sequence'):
                    value_dict['sequence'] = run_metadata[key].get('sequence') or run_metadata[key].get('seq')
                if cmd_kwargs.get('print_aSD_sequence'):
                    value_dict['anti-Shine-Dalgarno'] = run_metadata[key].get('anti-Shine-Dalgarno') or run_metadata[key].get('asd') or cmd_kwargs.get('aSD')
                result_set.append(value_dict)
            results.extend(result_set)
        

    ## Output - for all ways of running ##############################################################
    if outfile:
        #if the output is empty, we need to print a valid header...
        column_names = ['name', 'start_codon', 'start_position', 'expression', 'RBS_distance_bp', 'dG_total', 'dG_rRNA:mRNA', 'dG_mRNA', 'dG_spacing', 'dG_standby', 'dG_start_codon']
        if cmd_kwargs.get('print_mRNA_sequence'):
            column_names.insert(1, 'sequence')
        if cmd_kwargs.get('print_aSD_sequence'):
            column_names.insert(1, 'anti-Shine-Dalgarno')
        save_to_csv(column_names, results, outfile)
        if cmd_kwargs['verbosity'] > 0:
            print_string = f'Results written to {outfile}'
        if cmd_kwargs['verbosity'] > 0:
            if not rprint:
                print(print_string, file=sys.stdout)
            else:
                rprint(print_string)
    else:
        _print_output(results)


if __name__ == "__main__":
    main()
